IFNDEF _CSWMRG_
_CSWMRG_ equ 1

; --=====================================================================================--
; #CLASS:    CSWMRG 
; #VERSION:  1.0
; --=====================================================================================--
; Built by NaN's Object Class Creator
;  Sept 19, 2001
;
; By NaN ( jaymeson@hotmail.com )
; http://nan32asm.cjb.net
;
; --=====================================================================================--
; #AUTHOR: Smallwaves
; #DATE:   Dece. 27, 2001
;
; #DESCRIPTION:
;
;          The Single Writer/Multiple Reader Guard 

;          (<b>(SWMRG)</b>) ..
;
; --=====================================================================================--
; CLASS METHOD PROTOS
; --=====================================================================================--
   CSWMRG_Init    PROTO  :DWORD

; --=====================================================================================--
; FUNCTION POINTER PROTOS
; --=====================================================================================--
   CSWM_destructorPto    TYPEDEF  PROTO  :DWORD 
   CSWM_DonePto    TYPEDEF  PROTO  :DWORD
   CSWM_WaitToReadPto    TYPEDEF  PROTO  :DWORD
   CSWM_WaitToWritePto    TYPEDEF  PROTO  :DWORD

; --=====================================================================================--
; CLASS STRUCTURE
; --=====================================================================================--
   CLASS CSWMRG, CSWM
      CMETHOD destructor    ; MUST BE THE FIRST, OR OBJECTS.INC WILL FAIL
      CMETHOD Done    ; Call this when done accessing the resource
      CMETHOD WaitToRead    ; Call this to gain shared read access
      CMETHOD WaitToWrite    ; Call this to gain exclusive write access      
      m_hsemReaders     dd    ?
      m_hsemWriters     dd    ?
      m_nActive    SDWORD    ?
      m_nWaitingReaders   SDWORD    ?
      m_nWaitingWriters   SDWORD    ?
      m_cs CRITICAL_SECTION <>
   CSWMRG ENDS

.data

   BEGIN_INIT 
      dd offset CSWM_destructor_Funct
      dd offset CSWM_Done_Funct
      dd offset CSWM_WaitToRead_Funct
      dd offset CSWM_WaitToWrite_Funct
      dd 0
      dd 0
      SDWORD 0
      SDWORD 0
      SDWORD 0
      dd 0,0,0,0,0,0
   END_INIT

.code
; --=====================================================================================--
; #METHOD:      CONSTRUCTOR (NONE)
;
; #DESCRIPTION: Initially no readers want access, no writers want 
;				access, and no threads are accessing the resource..
;
; #RETURN:      Nothing..
; --=====================================================================================--
CSWMRG_Init  PROC uses edi esi lpTHIS:DWORD
   SET_CLASS CSWMRG
   SetObject edi, CSWMRG

    invoke CreateSemaphore,NULL, 0, MAXLONG, NULL
    mov [edi].m_hsemReaders,eax
    invoke CreateSemaphore,NULL, 0, MAXLONG, NULL
    mov [edi].m_hsemWriters,eax    
    invoke InitializeCriticalSection,addr [edi].m_cs
   ReleaseObject edi
   ret
CSWMRG_Init ENDP

; --=====================================================================================--
; #METHOD:      destructor (NONE)
;
; #DESCRIPTION: Do Some Clean work..
;
; #RETURN:      Nothing..
; --=====================================================================================--
CSWM_destructor_Funct  PROC uses edi lpTHIS:DWORD 
   SetObject edi, CSWMRG

	ifdef _DEBUG
   	 ;A SWMRG shouldn't be destroyed if any threads are using the resource
   		.if ([edi].m_nActive != 0)
		      invoke DebugBreak
		.endif
	endif

	mov [edi].m_nWaitingReaders,0
	mov [edi].m_nWaitingWriters,0
	mov [edi].m_nActive,0
	invoke DeleteCriticalSection,addr [edi].m_cs
	invoke CloseHandle,[edi].m_hsemReaders
	invoke CloseHandle,[edi].m_hsemWriters

   ReleaseObject edi
   ret
CSWM_destructor_Funct  ENDP

; --=====================================================================================--
; #METHOD:      Done (access)
;
; #DESCRIPTION: Read or Write has done..
;
; #RETURN:      Nothing..
; --=====================================================================================--
CSWM_Done_Funct  PROC uses edi lpTHIS:DWORD
local hsem:dword
local lCount:dword  
   SetObject edi, CSWMRG

	;Ensure exclusive access to the member variables
	invoke EnterCriticalSection,addr [edi].m_cs
		
	.if [edi].m_nActive > 0   
		;Readers have control so a reader must be done
		dec [edi].m_nActive
	.else
		;Writers have control so a writer must be done
		inc [edi].m_nActive
	.endif
	
	mov hsem,NULL;Assume no threads are waiting
	mov lCount,1;Assume only 1 waiter wakes; always true for writers
	
	.if [edi].m_nActive == 0
		;No thread has access, who should wake up?
		;NOTE: It is possible that readers could never get access
        ;       if there are always writers wanting to write
        
        .if [edi].m_nWaitingWriters > 0
        	;Writers are waiting and they take priority over readers
        	mov [edi].m_nActive,-1;A writer will get access
        	dec [edi].m_nWaitingWriters; One less writer will be waiting
        	push [edi].m_hsemWriters
        	pop hsem;Writers wait on this semaphore
        		;NOTE: The semaphore will release only 1 writer thread
        
        .elseif [edi].m_nWaitingReaders > 0
        	;Readers are waiting and no writers are waiting
        	push [edi].m_nWaitingReaders
        	pop [edi].m_nActive;All readers will get access
        	mov [edi].m_nWaitingReaders,0;No readers will be waiting
        	push [edi].m_hsemReaders
        	pop hsem;Readers wait on this semaphore
        	push [edi].m_nActive
        	pop lCount;Semaphore releases all readers
        	
        .else
        	;There are no threads waiting at all; no semaphore gets 
			;released
		.endif
	.endif
	
	;Allow other threads to attempt reading/writing
	invoke LeaveCriticalSection,addr [edi].m_cs
	
	.if hsem != NULL
		;Some threads are to be released
		invoke ReleaseSemaphore,hsem, lCount, NULL
	.endif
	
   ReleaseObject edi
   ret
CSWM_Done_Funct  ENDP

; --=====================================================================================--
; #METHOD:      WaitToRead ()
;
; #DESCRIPTION: Wait to read..
;
; #RETURN:      Nothing..
; --=====================================================================================--
CSWM_WaitToRead_Funct  PROC uses edi lpTHIS:DWORD
   SetObject edi, CSWMRG

	; Ensure exclusive access to the member variables
	invoke EnterCriticalSection,addr [edi].m_cs
	
	;Are there writers waiting or is a writer writing?
	.if [edi].m_nWaitingWriters || [edi].m_nActive < 0	
		mov eax,TRUE
		;This reader must wait, increment the count of waiting readers
		inc [edi].m_nWaitingReaders
	
	.else
		xor eax,eax	   
   		;This reader can read, increment the count of active readers
   		inc [edi].m_nActive
 	.endif
	push eax
	 	
 	; Allow other threads to attempt reading/writing
 	invoke  LeaveCriticalSection,addr [edi].m_cs
 	
 	pop eax
 	.if eax
 		;This thread must wait
 		invoke WaitForSingleObject,[edi].m_hsemReaders, INFINITE
 	.endif
   		
   ReleaseObject edi
   ret
CSWM_WaitToRead_Funct  ENDP

; --=====================================================================================--
; #METHOD:      WaitToWrite ()
;
; #DESCRIPTION: Wait to write..
;
; #RETURN:      Nothing..
; --=====================================================================================--
CSWM_WaitToWrite_Funct  PROC uses edi lpTHIS:DWORD
   SetObject edi, CSWMRG

	;Ensure exclusive access to the member variables
	invoke EnterCriticalSection,addr [edi].m_cs
	
	; Are there any threads accessing the resource?
	.if [edi].m_nActive != 0
		;This writer must wait, increment the count of waiting writers
		mov eax,TRUE
		inc [edi].m_nWaitingWriters
		
	.else
		;This writer can write, decrement the count of active writers
		mov [edi].m_nActive,-1
		xor eax,eax
	.endif
	push eax
		
	;Allow other threads to attempt reading/writing
	invoke LeaveCriticalSection,addr [edi].m_cs
	
	pop eax
	.if eax
		; This thread must wait
		invoke WaitForSingleObject,[edi].m_hsemWriters, INFINITE   
	.endif
	   
   ReleaseObject edi
   ret
CSWM_WaitToWrite_Funct  ENDP

ENDIF
